const path = require('path');
const chalk = require('chalk');
const fs = require('fs');
const uglify = require('uglify-js');
const browserify = require('browserify');
const prettyBytes = require('pretty-bytes');
const resolveEntry = require('../util/resolve-entry');
const getTransforms = require('../util/get-transforms');
const css = require('../plugins/css');
const resolverPlugin = require('../plugins/resolve');
const sketchPlugin = require('../plugins/sketch/plugin');
const { SimpleError } = require('../util/logger');
const getDefaultHTML = require('../util/get-default-html');
const mkdirp = require('mkdirp');
const fromString = require('from2-string');

module.exports = function (entryFile, opts = {}) {
  const logger = opts.logger;
  const cwd = opts.cwd;
  const baseDir = opts.dir;
  const file = resolveEntry(entryFile);
  opts = Object.assign({}, opts, {
    entry: file,
    mode: 'production'
  });
  opts.env = Object.assign({}, opts.env, {
    NODE_ENV: 'production'
  });
  opts.debug = Boolean(opts.debug);

  const output = opts.js;
  const hasCSS = opts.css;
  let outputFile;
  if (output && path.isAbsolute(output)) outputFile = output;
  else if (output) outputFile = path.resolve(baseDir, output);

  if (outputFile && path.resolve(outputFile) === path.resolve(file)) {
    throw new SimpleError(`The entry file, ${chalk.bold(path.relative(cwd, file))}, cannot be the same as the output file for bundling. Try specifying a different entry point or --js output.`);
  }

  const isWrite = opts.write !== false;
  const createDir = (dir) => {
    try {
      mkdirp.sync(dir);
    } catch (err) {
      logger.error(err);
    }
  };

  return getTransforms(opts).then(transform => {
    // See if we should auto-generate an HTML
    if (isWrite) {
      // Ensure the output folder exists
      createDir(baseDir);
      createDir(path.dirname(outputFile));

      // See if we should write a default HTML file
      if (opts.html && typeof opts.html === 'string') {
        const htmlFile = path.isAbsolute(opts.html) ? opts.html : path.resolve(baseDir, opts.html);
        if (!fs.existsSync(htmlFile)) {
          const htmlDoc = getDefaultHTML(opts);
          logger.bullet(`HTML → ${chalk.bold(path.relative(opts.cwd, htmlFile))}`);
          try {
            fs.writeFileSync(htmlFile, htmlDoc);
          } catch (err) {
            logger.error(err);
          }
        } else {
          logger.bullet(`HTML → Already exists in ${chalk.bold(path.relative(opts.cwd, baseDir))}, skipping`);
        }
      }
    }

    const entries = [];

    // If we want to polyfill for older browsers, and if we aren't optimize polyfill per-file,
    // ensure it is added as the root file
    if (opts.polyfill && !opts.optimize) {
      entries.push(require.resolve('../util/polyfills.js'));
    }

    if (opts.sketch) {
      entries.push(sketchPlugin.getEntry(opts));
    } else {
      entries.push(file);
    }

    const b = browserify(entries, Object.assign({}, opts.browserifyOpts, { debug: opts.debug }));
    const cssOptions = {
      write: isWrite,
      failOnError: true,
      url: opts.css,
      output: opts.css ? path.resolve(baseDir, opts.css) : false,
      compress: opts.minify,
      logger
    };

    // apply default transforms
    transform.forEach(t => b.transform(t));

    // Include exporter plugin to handle sketch mode
    b.plugin([ sketchPlugin, opts ]);

    // apply optimizations
    if (opts.optimize) {
      b.transform(require('unassertify'), { global: true });
      b.transform(require('unreachable-branch-transform'), { global: true });
    }

    // apply CSS plugin
    if (hasCSS) b.plugin([ css.plugin, cssOptions ]);
    b.plugin([ resolverPlugin, opts ]);

    // apply minificaiton related transforms
    if (opts.minify && opts.optimize) {
      if (!opts.fullPaths) {
        b.plugin(require('bundle-collapser/plugin'));
      }
      b.plugin(require('common-shakeify'));
    }

    return new Promise((resolve, reject) => {
      let cssResult;
      if (hasCSS) {
        b.once('css', result => {
          cssResult = result;
        });
      }
      b.bundle((err, src) => {
        if (err) return reject(err);
        if (opts.minify) {
          var min = uglify.minify(src.toString(), {
            mangle: true,
            compress: {
              dead_code: true,
              evaluate: true
            }
          });
          if (min.error) return reject(min.error);
          src = min.code;
        }

        src = `${(src || '').toString().trim()}\n`;

        const result = {
          code: src,
          css: cssResult
        };
        const done = () => {
          if (isWrite) {
            const bytes = chalk.dim(`${prettyBytes(src.length)}`);
            if (logger) logger.bullet(`JS   → ${chalk.bold(path.relative(cwd, outputFile))} ${bytes}`);
          }
          resolve(result);
        };
        if (isWrite) {
          if (opts.stdout) {
            process.stdout.write(src, done);
          } else {
            fs.writeFile(outputFile, src, err => {
              if (err) {
                return reject(err);
              } else {
                done();
              }
            });
          }
        } else {
          done();
        }
      });
    });
  });
};
