const path = require('path');
const getPackage = require('../util/get-package');
const inquirer = require('inquirer');
const templateCopy = require('../util/template-copy');
const parseJSON5 = require('../util/parse-json5');
const ignoreFiles = require('../util/ignore-files');
const fs = require('fs');
const npmCommand = require('../util/npm-command');
const ora = require('ora');
const chalk = require('chalk');
const version = require('../../package.json').version;
const detectInstalled = require('detect-installed');
const { SimpleError } = require('../util/logger');

const builtinTemplates = path.resolve(__dirname, '../../templates');
const templateSettingsFileName = '.texel.template';

const hasTexel = cmd => /texel/.test(cmd);

module.exports = function (entry, opts = {}) {
  const logger = opts.logger;
  const outputDir = opts.cwd;
  const force = opts.force;
  const isSpinner = true;
  const spinner = isSpinner ? ora('Reading package.json') : null;

  const folderName = path.basename(outputDir);
  logger.bullet(`Creating ${chalk.bold(folderName)}`);
  return (force ? Promise.resolve() : ensureSafeToCreate(opts))
    .then(() => {
      if (spinner) spinner.start();
      getPackage(outputDir);
    })
    .then(pkg => {
      return pkg;
    }).then(pkg => {
      // see if we can get a template settings from the directory
      const templateName = opts.template || 'default';
      const templateFolder = path.resolve(builtinTemplates, templateName);
      const templateSettingsFile = path.resolve(templateFolder, templateSettingsFileName);
      return parseJSON5(templateSettingsFile, { optional: true }).then(templateSettings => {
        return { pkg, templateSettings, templateFolder, templateName };
      });
    }).then(ev => {
      // Copy files
      const dirName = path.basename(outputDir);
      const data = {
        name: dirName,
        texelVersion: version,
        user: '',
        authorName: '',
        authorEmail: '',
        authorUrl: ''
      };
      return templateCopy(ev.templateFolder, outputDir, {
        data,
        ignore: [ templateSettingsFileName ]
      }).then(() => ev);
    })
    .then(ev => {
      return getPackage(outputDir).then(pkg => {
        return Object.assign({}, ev, { pkg });
      });
    })
    // First run a clean install to get any package.json defined dependencies
    .then(ev => {
      // No package.json, just skip installing defaults
      if (!ev.pkg) return ev;
      return runInstall('Installing dependencies', ev);
    })
    // Then run a new --save install if necessary
    .then(ev => {
      const install = ev.templateSettings.install ? ev.templateSettings.install : {};
      const deps = (install && install.dependencies) || [];
      return runInstall('Installing additional dependencies', ev, deps, '--save');
    })
    // Finally run a --save-dev, devDeps if necessary
    .then(ev => {
      const install = ev.templateSettings.install ? ev.templateSettings.install : {};
      const deps = (install && install.devDependencies) || [];
      return runInstall('Installation development dependencies', ev, deps, '--save-dev');
    })
    .then(ev => {
      if (spinner) {
        spinner.stop();
        spinner.clear();
        logger.bullet('Done ✨');
      }
      let startScript, bundleScript;

      if (ev.pkg && ev.pkg.scripts) {
        const runScript = ev.pkg.scripts;
        if (runScript.start && hasTexel(runScript.start)) {
          startScript = 'npm run start';
        }
        if (runScript.bundle && hasTexel(runScript.bundle)) {
          bundleScript = 'npm run bundle';
        }
      }

      if (startScript && bundleScript) {
        logger.wrap(`To start developing, run:\n  ${startScript}\n\nOr, to bundle your website:\n  ${bundleScript}`);
      } else {
        // Not using local npm scripts, see if we should tell user to install it globally
        return detectInstalled('texel').then(installed => {
          startScript = 'texel --open';
          bundleScript = 'texel bundle';
          let message = '';
          if (!installed) {
            message = `🚨  ${chalk.red('Heads up!')} It looks like ${chalk.green(chalk.bold('texel'))} hasn't been installed globally.\n\nInstall it globally with the following command:\n  npm install texel -g\n\nThen, to start developing, run:\n  ${startScript}\n\nOr, to bundle your website:\n  ${bundleScript}\n\n`;
          } else {
            message = `To start developing, run:\n  ${startScript}\n\nOr, to bundle your website:\n  ${bundleScript}\n\n`;
          }
          logger.wrap(message);
        });
      }
    })
    .catch(err => {
      if (spinner) {
        spinner.stop();
        if (!err.isSimpleError) spinner.fail(`Something went wrong!`);
      }
      throw err;
    });

  function runInstall (status, ev, deps = [], type = '') {
    if (spinner) spinner.text = status;
    const isSkip = type && deps.length === 0;
    return (isSkip
      ? Promise.resolve()
      : npmCommand('install', type ? deps.concat([ type ]) : deps, { cwd: outputDir })
    ).then(() => ev);
  }
};

function ensureSafeToCreate (opts) {
  const cwd = opts.cwd;
  const ignorePrompt = opts.ignorePrompt;

  return isFolderEmpty(cwd).then(empty => {
    if (!empty) {
      if (ignorePrompt) {
        throw new SimpleError('Your current folder is not empty. Try again on an empty folder, or use the --force option.');
      } else {
        return prompt(`🚨  ${chalk.red('WARNING: Your current folder is not empty.')}\n  Are you sure you want to overwrite its contents with a new project?`)
          .then(answer => {
            if (!answer) throw new SimpleError(`Cancelled. Try again in an empty folder, or use the --force option.`);
          });
      }
    }
  });
}

function isFolderEmpty (dir) {
  return new Promise((resolve, reject) => {
    fs.readdir(dir, (err, files) => {
      if (err) return reject(err);
      files = files.filter(f => {
        return !ignoreFiles.some(reg => reg.test(f));
      });
      resolve(files.length === 0);
    });
  });
}

function prompt (message) {
  return inquirer.prompt([ {
    default: false,
    type: 'confirm',
    name: 'confirm',
    message: message
  } ]).then(answers => {
    return answers.confirm;
  });
}

// .then(undefined, () => {
//   if (!force) {
//     throw new SimpleError(`It looks like you already have a ${} in this folder.\nTry running ${chalk.bold('texel create')} on an empty project folder.`);
//   }
// })

// throw new SimpleError(`It looks like you already have files in this folder.\nTry running ${chalk.bold('texel create')} on an empty project folder.`);
