const budo = require('budo');
const clientErrorHandler = require('budo/lib/error-handler').render;
const getTransforms = require('../util/get-transforms');
const createDevtoolMiddleware = require('../plugins/devtool-middleware');
const sketchPlugin = require('../plugins/sketch/plugin')
const path = require('path');
const bodyParser = require('body-parser');
const chalk = require('chalk');
const cssUtil = require('../plugins/css');
const stripAnsi = require('strip-ansi');
const shaderReloader = require('shader-reload/server');
const resolveEntry = require('../util/resolve-entry');
const getDefaultHTML = require('../util/get-default-html');
const fromString = require('from2-string');
const rightNow = require('right-now');
const prettyMs = require('pretty-ms');

module.exports = function (entryFile, opts = {}) {
  if (!process.env.NODE_ENV) {
    process.env.NODE_ENV = 'development';
  }

  const logger = opts.logger;
  const cwd = opts.cwd;
  const baseDir = opts.dir;

  const output = opts.js;
  const entry = resolveEntry(entryFile);
  opts = Object.assign({}, opts, { entry, mode: 'development' });
  opts.debug = opts.debug !== false;

  if (output && path.resolve(output) === path.resolve(entry)) {
    logger.warn(`The input file, ${chalk.bold(path.relative(cwd, entry))}, is the same as the --js output file. This may cause problems when you plan to bundle your site. Instead, you should rename your entry file, or choose a different --js output file. Example:
  texel start index.js --js bundle.js`);
  }

  // See if we should auto-generate an HTML
  const isWrite = opts.write !== false;

  return getTransforms(opts).then(transform => {
    const hasCSS = Boolean(opts.css);
    const cssOptions = {
      write: isWrite,
      url: opts.css,
      output: opts.css ? path.resolve(baseDir, opts.css) : false,
      compress: false
    };

    const css = cssUtil.dev(cssOptions);

    let app;
    let currentDeps;
    let wss;
    let time = rightNow();
    let hasError = false;
    let isErrorState = false;

    let lastCSSError;
    let lastShaderError;

    const showErrorPopup = (err) => {
      if (!app) return;
      const msg = stripAnsi(err.toString());
      if (msg) {
        app.error(msg);
      }
    };

    const clearError = () => {
      if (!app) return;
      app.error(false);
    };

    const onCSSError = err => {
      if (!hasCSS) return;
      lastCSSError = err;
      logger.statusError(err.toString());
      showErrorPopup(err);
    };

    const onCSSChange = () => {
      if (!hasCSS) return;
      if (app) {
        app.reload(cssOptions.url);
      }
    };

    css.on('start', () => {
      if (!hasCSS) return;
      if (lastCSSError && !lastShaderError) {
        // Clear popup since we were showing an error message
        clearError();
      }
      lastCSSError = null;
      css.removeListener('error', onCSSError);
      css.removeListener('finish', onCSSChange);
      css.once('error', onCSSError);
      css.once('finish', onCSSChange);
    });

    const devtool = createDevtoolMiddleware(opts);
    const defaultDoc = getDefaultHTML(opts);

    const entries = [];

    // include polyfill if necessary first...
    if (opts.polyfill) entries.push(require.resolve('../util/polyfills.js'));

    if (opts.sketch) {
      entries.push(sketchPlugin.getEntry(opts));
    } else {
      entries.push(entry);
    }

    app = budo(entries, {
      title: 'texel',
      debug: opts.debug,
      dir: baseDir,
      pushstate: opts.pushstate,
      stream: null,
      live: false,
      open: opts.open,
      serve: output,
      port: parseInt(opts.port, 10),
      middleware: [
        bodyParser.json({
          limit: '2gb'
        }),
        devtool,
        css.middleware
      ],
      errorHandler,
      defaultIndex: () => {
        return fromString(defaultDoc);
      },
      browserify: Object.assign({}, opts.browserifyOpts, {
        plugin: [
          [ sketchPlugin, opts ],
          css.plugin,
          [ require('../plugins/resolve.js'), opts ]
        ],
        transform: transform.concat([
          require('shader-reload/transform')
        ])
      })
    });

    const broadcast = (clients, message) => {
      clients.forEach(client => {
        if (client.readyState === client.OPEN) {
          client.send(message);
        }
      });
    };

    app
      .live()
      .watch()
      .on('bundle-error', (err) => {
        hasError = true;
        isErrorState = true;
        if (shaderReloader.isShaderReload(currentDeps)) {
          lastShaderError = err;
          if (!lastCSSError) showErrorPopup(lastShaderError);
        }
        logger.statusError(err.toString());
      })
      .on('pending', (deps) => {
        hasError = false;
        currentDeps = deps;
        time = rightNow();
        logger.status('Bundling');

        if (wss && shaderReloader.isShaderReload(deps)) {
          // Skip shader reload events in pending, and send the
          // new upate in 'update' (after bundle finishes)
        } else if (deps.some(file => !/\.(css|less|sass|scss)$/i.test(file))) {
          // Force a browser reload for JS assets
          app.reload();
        }
      })
      .on('update', (src, deps) => {
        const dt = rightNow() - time;
        const delta = prettyMs(dt);
        let status;
        if (wss && shaderReloader.isShaderReload(deps)) {
          // Shader reload event, send the message data
          const event = JSON.stringify(shaderReloader.getEvent(deps));
          broadcast(wss.clients, event);
          status = 'Updated GLSL shaders in';
        } else {
          status = 'Bundled in';
        }

        if (isErrorState && !hasError) {
          isErrorState = false;

          if (lastShaderError && !lastCSSError) {
            clearError();
          }
          lastShaderError = null;
          logger.clearStatusError();
          logger.status(`Error resolved`);
        } else if (!hasError && !lastCSSError && !lastShaderError) {
          logger.status(`${status} ${chalk.bold(delta)}`);
        }
      })
      .on('watch', (ev, file) => {
        if (cssOptions.output && path.resolve(file) === path.resolve(cssOptions.output)) {
          // Ignore the CSS file on disk since we are serving it in-memory
          return;
        }
        // TODO: ignore required() CSS but still reload link'd CSS
        // If non-shader file, reload resource
        if (!/\.shader\.js$/i.test(file)) {
          app.reload(file);
        }
      });

    return new Promise((resolve, reject) => {
      app
        .once('error', reject)
        .once('connect', ev => {
          logger.bullet(`Development server is running on:\n    ${chalk.magenta(`http://${ev.host}:${ev.port}/`)}\n    ${chalk.magenta(`http://localhost:${ev.port}/`)}`);
          logger.bullet(`Hit ${chalk.bold('Ctrl-C')} to stop the server`);

          wss = ev.webSocketServer;
          if (wss) {
            wss.on('connection', () => {
              const lastErr = lastCSSError || lastShaderError;
              if (lastErr) showErrorPopup(lastErr);
            });
          }

          resolve(Object.assign({}, app, ev, { url: ev.uri }));
        });
    });
  });

  function errorHandler (err, cwd, rootDirName) {
    var msgStr = stripAnsi(err.message);
    var params = [
      JSON.stringify(msgStr),
      JSON.stringify(cwd),
      JSON.stringify(rootDirName)
    ].join(',');
    return ';(' + clientErrorHandler + ')(' + params + ');';
  }
};
