const chalk = require('chalk');
const parseArgs = require('./util/parse-args');
const getBaseDir = require('./util/get-base-dir');
const logger = require('./util/logger');
const defined = require('defined');

const availableCommands = [
  'start', 'bundle', 'help', 'clean', 'create'
];

let prettyLoggerAttached = false;

function attachPrettyError (log) {
  if (prettyLoggerAttached) return;
  prettyLoggerAttached = true;
  process.on('uncaughtException', errListener);
  process.on('unhandledRejection', errListener);

  function errListener (err) {
    console.error(err);
    process.exit(1);
  }
}

module.exports = function (commands = [], entry = '', opts = {}) {
  const trySome = `Try some of the following:
  texel help
  texel create
  texel start
  texel bundle`;

  commands = [].concat(commands).filter(Boolean);
  // default to start command
  if (!commands || commands.length === 0) {
    commands.push('start');
  }

  // setup defaults
  opts = Object.assign({}, opts);
  opts.port = parseInt(defined(opts.port, 9966), 10);
  opts.js = opts.js === false ? false : (opts.js || 'bundle.js');
  opts.css = opts.css === false ? false : (opts.css || 'bundle.css');
  opts.html = opts.html === false ? false : (opts.html || 'index.html');
  opts.minify = opts.minify !== false;
  opts.quiet = Boolean(opts.quiet);
  opts.cwd = process.cwd();
  opts.open = Boolean(opts.open);
  opts.dir = getBaseDir(defined(opts.dir, 'public'), { cwd: opts.cwd });
  opts.logger = logger(opts);
  opts.env = Object.assign({}, process.env, opts.env, { NODE_ENV: 'development' });
  opts._noParseFiles = [
    'node_modules/three/build/three.js',
    'node_modules/three/build/three.min.js'
  ];
  opts.browserifyOpts = {
    // Speed up ThreeJS apps.
    // Perhaps this is too magical, may be removed later.
    noParse: opts._noParseFiles,
    fullPaths: opts.fullPaths,
    bare: opts.bare,
    node: opts.node
  };

  if (!opts.js) throw new logger.SimpleError(`Oops, it looks like you specified --no-js or an empty path for the output.\n\nTry specifying a path, like this:\n  texel [command] index.js`);

  const start = (command = '') => {
    opts.logger.bullet(chalk.green(`texel ${chalk.bold(command)}`));
    switch (command) {
      case 'start': return require('./commands/start')(entry, opts);
      case 'bundle': return require('./commands/bundle')(entry, opts);
      case 'clean': return require('./commands/clean')(entry, opts);
      case 'create': return require('./commands/create')(entry, opts);
      default:
        throw new logger.SimpleError(`Unknown texel command: ${chalk.bold(command)}\n\n${trySome}`.trim());
    }
  };

  return commands.reduce((p, cmd) => {
    return p.then(result => start(cmd));
  }, Promise.resolve());
};

module.exports.cli = function (args = []) {
  // Parse out commands
  const commands = [];
  const origArgs = args.slice();
  for (let i = 0; i < origArgs.length; i++) {
    const arg = origArgs[i];
    if (availableCommands.includes(arg)) {
      // Still a command
      args = origArgs.slice(i + 1);
      commands.push(arg);
    } else {
      // No longer at any commands
      break;
    }
  }

  const opts = parseArgs(args);
  const entry = opts._[0];
  delete opts._;

  const log = logger(opts);
  attachPrettyError(log);

  return module.exports(commands, entry, opts).catch(err => {
    log.error(err);
    process.exit(1);
  });
};
