const prettyBytes = require('pretty-bytes');
const path = require('path');
const chalk = require('chalk');
const postcss = require('postcss');
const URL = require('url');
const cleanCSS = require('postcss-clean');
const cssnext = require('postcss-cssnext');
const postCSSImport = require('postcss-import');
const through = require('through2');
const fs = require('fs');
const detective = require('detective');
const { EventEmitter } = require('events');
const { SimpleError } = require('../../util/logger');
const transformSASS = require('./transform-sass');
const transformLESS = require('./transform-less');
const cssRegex = /\.(css|less|sass|scss)$/i;

module.exports.plugin = function (bundler, opt = {}) {
  const basedir = opt.basedir || process.cwd();
  let outputFile = opt.output;
  let isVirtual = false;
  const isWrite = opt.write !== false;
  // User specified a path for output
  if (typeof outputFile === 'string') {
    outputFile = path.isAbsolute(outputFile) ? outputFile : path.resolve(basedir, outputFile);
  } else if (typeof outputFile.emit === 'function') {
    isVirtual = true;
  }

  // Use a dummy file name when processing postcss in virtual mode
  const defaultFileName = 'main.css';
  const outputFileName = (isVirtual ? defaultFileName : outputFile) || defaultFileName;

  const logger = opt.logger;
  const onComplete = opt.onComplete;
  const isFailOnError = opt.failOnError;

  const plugins = [].concat(opt.plugins).filter(Boolean);
  const basePlugins = [
    postCSSImport(),
    cssnext(),
    opt.compress ? cleanCSS() : undefined
  ].filter(Boolean);
  const cssParser = postcss(plugins.concat(basePlugins));
  const emptyModule = 'module.exports = undefined;';

  let allDeps = [];
  bundler.on('reset', addHooks);

  let firstRecord = true;
  bundler.pipeline.get('record').unshift(through.obj(function (chunk, enc, next) {
    if (firstRecord) {
      firstRecord = false;
      // First, cache all CSS source code, and also transform it to plain empty JS module
      // so that other transforms will not try to parse it
      this.push({
        transform: transform,
        global: undefined
      });
      this.push({
        transform: transform,
        global: true
      });
      next(null, chunk);
    } else {
      next(null, chunk);
    }
  }));

  addHooks();

  function addHooks () {
    allDeps.length = 0;

    if (isVirtual) {
      outputFile.emit('start');
    }

    bundler.pipeline.get('unbom').push(through.obj(function (chunk, enc, next) {
      if (cssRegex.test(chunk.file)) {
        transformCSSByFile(chunk.file, bundler)
          .then(result => {
            next(null, Object.assign({}, chunk, { source: emptyModule, css: result }));
          }, err => {
            // If we aren't failing, just log to console to keep things clear
            // for the developer.
            if (!isFailOnError && logger) {
              logger.statusError(err.toString());
            }
            if (isVirtual) {
              outputFile.emit('error', err);
            }
            if (isFailOnError) next(new Error(err.toString()));
            else next(null, Object.assign({}, chunk, { source: emptyModule, css: `/* ${err.message.replace(/(\r\n|\n|\r)/gm, '')} */` }));
          });
      } else {
        next(null, chunk);
      }
    }));

    bundler.pipeline.get('emit-deps').push(through.obj((chunk, enc, next) => {
      allDeps.push(chunk);
      next(null, chunk);
    }));

    bundler.pipeline.get('wrap').push(writer());
  }

  function writer () {
    return through((chunk, enc, next) => next(null, chunk), (next) => {
      const cssDeps = allDeps.filter(dep => Boolean(dep.css));

      // Small perf boost: we can skip ordering if there is just a single dep
      const ordered = cssDeps.length > 1 ? getOrderedCSS(allDeps) : cssDeps;

      // For convenience, but may be helpful?
      const cssFiles = ordered.map(d => d.file);
      if (!isVirtual && outputFile && cssFiles.some(file => file === outputFile)) {
        const errMessage = `Cannot write to ${chalk.bold(path.relative(process.cwd(), outputFile))} as you've imported it in your JavaScript. Perhaps you can try specify a different --css output file, or avoid importing the same file you plan to output.`;
        return next(new SimpleError(errMessage));
      }

      // Useful to warn user that no CSS option is specified
      if (!outputFile && cssFiles.length > 0 && logger) {
        logger.info(`It looks like you are importing CSS files into your JavaScript, but haven't specified a --css output file. Try the following:
  texel [command] [entry] --css bundle.css`);
      }

      const css = ordered.map(d => d.css).filter(Boolean).join('\n');

      bundler.emit('css', {
        code: css || '',
        modules: cssFiles
      });

      if (isVirtual) {
        outputFile.emit('finish', { css });
        next(null);
      } else if (isWrite && outputFile) {
        fs.writeFile(outputFile, css, err => {
          if (err) {
            return next(err);
          } else {
            const bytes = chalk.dim(`${prettyBytes(css.length)}`);
            if (logger) logger.bullet(`CSS  → ${chalk.bold(path.relative(basedir, outputFile))} ${bytes}`);
            next(null);
          }
        });
      } else {
        next(null);
      }
    });
  }

  function getOrderedCSS (deps) {
    const moduleMap = deps.reduce((dict, dep) => {
      dict[dep.id] = Object.assign({}, dep);
      return dict;
    }, {});

    const cssIds = [];
    for (let i = 0; i < deps.length; i++) {
      if (deps[i].css) cssIds.push(deps[i].id);
    }

    const finalCSSOrder = [];
    const filesChecked = [];
    const walk = item => {
      // Avoid circular deps / endless walking
      if (filesChecked.includes(item.file)) return;
      filesChecked.push(item.file);

      const originalDeps = item.deps;
      // Get the proper order of dependencies here
      let itemDeps = Object.keys(originalDeps).map(key => {
        return {
          name: key,
          id: originalDeps[key]
        };
      });

      // If it has dependencies, parse AST to get actual order
      let dependencyOrder = item.dependencyOrder;
      if (!dependencyOrder) {
        // cache it for this module
        dependencyOrder = itemDeps.length > 0 ? detective(item.source) : [];
        // ensure names are strings, in case we are using bundle-collapser
        dependencyOrder = dependencyOrder.map(n => String(n));
        item.dependencyOrder = dependencyOrder;
      }

      // No dependencies in this module, we can bail early
      if (dependencyOrder.length === 0) return;

      itemDeps.forEach(dep => {
        dep.order = dependencyOrder.indexOf(dep.name);
      });

      // Ignore any deps not matching our require statements
      itemDeps = itemDeps.filter(dep => dep.order >= 0);

      // Sort by order in actual AST
      itemDeps.sort((a, b) => a.order - b.order);

      // Now walk through each of those
      itemDeps.forEach(dep => {
        const other = moduleMap[dep.id];
        if (!other) return;
        if (other.css) {
          // If it's already found, remove it from the stack
          const idx = finalCSSOrder.indexOf(dep.id);
          if (idx !== -1) finalCSSOrder.splice(idx, 1);

          // Push it back on to the end
          finalCSSOrder.push(dep.id);
        } else {
          // Other dep is not a CSS dep, we need to walk it
          walk(other);
        }
      });
    };

    // Now walk entire tree from entry point to determine
    // the correct ordering of CSS dependencies
    const entries = deps.filter(dep => dep.entry).sort((a, b) => a.order - b.order);
    entries.forEach(entry => {
      walk(Object.assign({}, entry));
    });

    // Return the CSS dependencies
    const finalCSSDeps = finalCSSOrder.map(id => moduleMap[id]);
    return finalCSSDeps;
  }

  function transformCSSByFile (file, compiler) {
    const ext = path.extname(file).toLowerCase();

    const styleOpt = Object.assign({}, opt, {
      outputFileName,
      outputFile,
      logger,
      isFailOnError,
      plugins,
      basedir,
      compiler
    });

    return new Promise((resolve, reject) => {
      fs.readFile(file, 'utf-8', (err, data) => {
        if (err) reject(err);
        else resolve(data);
      });
    }).then(body => {
      switch (ext) {
        case '.css':
          return transformCSS(body, file, styleOpt);
        case '.scss':
        case '.sass':
          return transformSASS(body, file, styleOpt);
        case '.less':
          return transformLESS(body, file, styleOpt);
        default:
          throw new SimpleError(`Invalid style sheet extension ${ext}`);
      }
    });
  }

  function transformCSS (body, file, styleOpt) {
    return cssParser.process(body, {
      failOnError: true,
      map: false,
      from: file,
      to: outputFileName
    }).then(result => {
      result.messages.forEach(msg => {
        if (msg.file && msg.type === 'dependency') {
          styleOpt.compiler.emit('file', msg.file);
        }
      });
      return result.css;
    });
  }

  function transform (file, opt) {
    if (!cssRegex.test(file)) return through();
    return through((chunk, enc, next) => {
      next(null);
    }, function (next) {
      this.push(emptyModule);
      next();
    });
  }
};

module.exports.dev = function (opt) {
  const outputFile = opt.output;
  const isServingCSS = Boolean(outputFile);
  let cssUrl = opt.url;
  // default to output filename
  if (outputFile && !cssUrl) cssUrl = path.basename(outputFile);
  // ensure it starts with a slash
  cssUrl = !/^[\\/.]/.test(cssUrl) ? `/${cssUrl}` : cssUrl;

  const emitter = new EventEmitter();
  let lastCSS = '';
  let ready = false;
  emitter.on('start', () => {
    ready = false;
  });
  emitter.on('finish', result => {
    ready = true;
    lastCSS = result && result.css;
  });
  const onError = () => {
    ready = true;
    lastCSS = '';
  };
  emitter.on('bundle-error', onError);
  emitter.on('error', onError);
  const done = (res, next) => {
    res.statusCode = 200;
    res.setHeader('Content-Type', 'text/css');
    res.end(lastCSS);
  };
  const middleware = (req, res, next) => {
    if (isServingCSS && URL.parse(req.url).pathname === cssUrl) {
      if (ready) { // pipeline already finished
        done(res);
      } else { // pipeline still in progress
        const handleNext = () => {
          done(res);
          emitter.removeListener('finish', handleNext);
          emitter.removeListener('error', handleNext);
          emitter.removeListener('bundle-error', handleNext);
        };
        emitter.once('finish', handleNext);
        emitter.once('error', handleNext);
        emitter.once('bundle-error', handleNext);
      }
    } else {
      next(null);
    }
  };
  emitter.plugin = (bundler) => {
    const pOpts = Object.assign({}, opt, { output: outputFile ? emitter : outputFile });
    bundler.plugin(module.exports.plugin, pOpts);

    // patch bundle error so we can ensure CSS middleware completes
    const bundleFn = bundler.bundle;
    bundler.bundle = function () {
      // console.profile('bundle');
      const pipeline = bundleFn.apply(bundler, Array.prototype.slice.call(arguments));

      let onBundleErr;
      let detachListeners = () => {
        // console.profileEnd('bundle');
        pipeline.removeListener('error', onBundleErr);
        pipeline.removeListener('end', detachListeners);
      };

      onBundleErr = (err) => {
        emitter.emit('bundle-error', err);
        pipeline.removeListener('end', detachListeners);
      };

      pipeline.once('error', onBundleErr);
      pipeline.once('end', detachListeners);
      return pipeline;
    };
  };
  emitter.middleware = middleware;
  return emitter;
};
