const fs = require('fs');
const chalk = require('chalk');
const path = require('path');
const mkdirp = require('mkdirp');
const padLeft = require('pad-left');
const GifEncoder = require('gif-encoder');
const PngDecoder = require('png-stream/decoder');
const concatFrames = require('concat-frames');

module.exports = function (opts = {}) {
  const logger = opts.logger;
  const outDir = opts.cwd;

  let gif;
  let gifFrame = 0;
  let gifPath = null;

  return function (req, res, next) {
    if (req.method !== 'POST') return next(null);

    switch (req.url) {
      case '/texel/export/saveFile': return saveFile(req, res);
      case '/texel/export/beginRecord': return beginRecord(req, res);
      case '/texel/export/recordFrame': return recordFrame(req, res);
      case '/texel/export/endRecord': return endRecord(req, res);
      default: return next(null);
    }
  };

  function writeError (res, message, statusCode = 400) {
    res.writeHead(statusCode, message);
    res.end();
  }

  function finish (res, message = 'ok') {
    res.writeHead(200, message);
    res.end();
  }

  function getFile (req, write) {
    const fileName = resolveFile(req.body);
    const filePath = path.resolve(outDir, fileName);
    const fileDir = path.dirname(filePath);
    const next = () => {
      write({ fileName, filePath, fileDir });
    };
    if (fileDir !== outDir) {
      mkdirp(fileDir, err => {
        if (err) return logger.error(err);
        next();
      });
    } else next();
  }

  function saveFile (req, res) {
    if (!req.body || !req.body.data) {
      return writeError(res, 'No base64 data was provided to saveFile function');
    }

    const data = Buffer.from(req.body.data, 'base64');
    getFile(req, ({ filePath }) => {
      fs.writeFile(filePath, data, function (err) {
        if (err) {
          logger.error(err);
          return writeError(res, err.message);
        }

        // TODO: Make this persistent or not?
        logger.status(`Saved ${getSaveType(path.extname(filePath))} to ${chalk.bold(path.relative(outDir, filePath))}`, false, 'green');
        finish(res);
      });
    });
  }

  function beginRecord (req, res) {
    if (!req.body) {
      return writeError(res, 'No body was provided to beginRecord function');
    }

    if (gif) { // Clear previous
      gif.finish();
      gif = null;
    }

    const width = req.body.width;
    const height = req.body.height;
    if (typeof width !== 'number' || typeof height !== 'number') {
      return writeError(res, 'Must provide width and height options to beginRecord');
    }

    getFile(req, ({ filePath }) => {
      gifFrame = 0;
      gifPath = filePath;
      gif = new GifEncoder(width, height);
      if (typeof req.body.fps === 'number') {
        gif.setFrameRate(req.body.fps);
      }
      if (isFinite(req.body.loop)) gif.setRepeat(req.body.loop);
      if (isFinite(req.body.quality)) gif.setQuality(req.body.quality);

      const fileStream = fs.createWriteStream(filePath);

      // Start writing to output
      gif.pipe(fileStream);
      gif.writeHeader();

      logger.status(`Starting Recording at ${chalk.bold(path.relative(outDir, filePath))}`, false, 'green');
      finish(res);
    });
  }

  function recordFrame (req, res) {
    if (!req.body || !req.body.data) {
      return writeError(res, 'No base64 data was provided to recordFrame function');
    }
    if (!gif) {
      return writeError(res, 'Not yet encoding a GIF - make sure to call beginRecord() before recording frames');
    }
    const decoder = new PngDecoder();
    decoder.once('error', err => writeError(res, err.message));
    decoder.pipe(concatFrames(frames => {
      if (frames.length !== 1) return writeError(res, 'PNG base64 data did not have a valid number of frames (1)');
      const frame = frames[0];
      if (frame.width !== gif.width || frame.height !== gif.height) return writeError(res, 'PNG base64 data width/height did not match GIF width/height in beginRecord()');
      // Add to GIF
      gif.addFrame(frame.pixels);
      logger.status(`Recorded frame ${gifFrame++}`, false, 'green');
      // Finish frame recording
      finish(res);
    }));
    decoder.end(Buffer.from(req.body.data, 'base64'));
  }

  function endRecord (req, res) {
    if (!gif) {
      return writeError(res, 'Not yet encoding a GIF - make sure to call beginRecord() before recording frames');
    }
    gif.finish();
    gif = null;
    gifFrame = 0;
    logger.status(`Recorded GIF to ${chalk.bold(path.relative(outDir, gifPath))}`, false, 'green');
    finish(res);
  }

  function resolveFile (body) {
    const output = body.output || '';
    const file = body.file;

    let frame = null;
    let extension;
    if (typeof body.extension === 'string') extension = body.extension;
    else if (file) extension = path.extname(file);

    if (typeof body.frame === 'number') {
      let totalFrames;
      if (typeof body.totalFrames === 'number') {
        totalFrames = body.totalFrames;
      } else {
        totalFrames = (Math.max(10, body.frame) * 10);
      }
      frame = padLeft(String(body.frame), String(totalFrames).length, '0');
    }

    const formattedFile = formatFile(output, file, extension, frame);
    if (frame != null) {
      if (file) return formattedFile;
      return path.join(output, `${frame}${extension}`);
    } else {
      if (file) return formattedFile;
      // TODO: Could improve situations with multiple simultaneous saves (within same 1 second)
      // by adding a number to make each save more unique. Or, in docs just encourage users to use 'frame'
      // option for that sort of accuracy.
      else return path.join(output, defaultFile(extension));
    }
  }

  function formatFile (output = '', file = '', extension = '', frame = 0) {
    return path.join(output, file.replace(/%d/, frame).replace(/%s/, extension));
  }

  function getSaveType (ext) {
    let type = 'File';
    if (/\.(jpe?g|png|webp)/i.test(ext)) type = 'Image';
    else if (/\.svg/i.test(ext)) type = 'SVG';
    return type;
  }

  function defaultFile (ext) {
    const type = getSaveType(ext);
    const str = `${type} - ${new Date().toLocaleDateString()} at ${new Date().toLocaleTimeString()}${ext}`;
    return str.replace(/\//g, '-').replace(/:/g, '.');
  }
};
