const through = require('through2');
const stripAnsi = require('strip-ansi');
const chalk = require('chalk');
const path = require('path');
const fs = require('fs');
const { SimpleError } = require('../../util/logger');
const { EventEmitter } = require('events');

const entryFile = require.resolve('./default-entry.js');

const presets = {
  '2d': require.resolve('../presets/2d.js'),
  'webgl': require.resolve('../presets/webgl.js')
};

module.exports = function (bundler, opts = {}) {
  const extractExports = opts.extractExports.exports;
  const filesToWatch = [ opts.entry, entryFile ];

  let entryReady = false;
  let hasGeneratedEntry = false;
  let entryReadyEmitter = new EventEmitter();
  bundler.on('transform', function (tr, mfile) {
    tr.on('babelify-start', function (file) {
      if (file === opts.entry) {
        entryReady = false;
        extractExports.settings = {};
      }
    });
    tr.on('babelify-finish', function (file) {
      if (file === opts.entry) {
        entryReady = true;
        entryReadyEmitter.emit('ready');
      }
    });
  });

  bundler.exclude('texel/_plugin/entry');
  patchFileWatch();
  bundler.transform(transform);

  addHooks();
  bundler.on('reset', addHooks);

  function addHooks () {
    // If an entry wasn't generated (i.e. no need for a preset),
    // then we can safely strip it out of the deps so that our final
    // bundle won't even include it
    bundler.pipeline.get('deps').push(through.obj(function (chunk, enc, next) {
      if (!hasGeneratedEntry && chunk.file === entryFile) {
        return next(null);
      }

      if (!hasGeneratedEntry && chunk.entry && typeof chunk.order === 'number') {
        chunk = Object.assign({}, chunk, { order: chunk.order - 1 });
      }
      next(null, chunk);
    }));
  }

  function transform (file) {
    if (file !== entryFile) return through();
    let code = '';

    return through((chunk, enc, next) => {
      code += chunk.toString();
      next(null, '');
    }, function (next) {
      const done = () => {
        code = getEntryCode(this, code);
        this.push(code);
        next(null);
      };
      if (entryReady) done();
      else {
        entryReadyEmitter.once('ready', done);
      }
    });
  }

  function getEntryCode (stream, code) {
    let presetFile;
    if (extractExports.settings.confident === false) {
      opts.logger.warn(`It looks like you tried exporting settings for your sketch file at:\n  ${chalk.bold(path.relative(opts.cwd, opts.entry))}\n\nHowever, your JavaScript includes dynamic variables and cannot be analyzed at bundle time.

Incorrect example:

  const width = window.innerWidth;
  export const settings = { size: [ width, 20 ] };

Correct example:

  export const settings = { size: [ 20, 20 ] };`);
    } else if (extractExports.settings.confident === true && extractExports.settings.value) {
      const settings = extractExports.settings.value;
      const preset = settings.preset;
      if (preset) {
        if (preset in presets) { // built-in preset name
          presetFile = presets[preset];
        } else if (path.isAbsolute(preset)) { // require.resolve was inlined into path
          presetFile = preset;
        } else {
          const msg = `Your sketch preset has been set to "${chalk.bold(preset)}", but there is no built-in Texel preset by this name. Change the settings in your entry file at ${chalk.bold(path.relative(opts.cwd, opts.entry))}\n\nSupported built-in presets:\n  ${Object.keys(presets).join(', ')}\n\n`;
          stream.emit('error', new SimpleError(msg));
          hasGeneratedEntry = true;
          return `/** Bundle error */\nconsole.error(${stripAnsi(msg)});`;
        }
      }
    }

    if (presetFile) {
      hasGeneratedEntry = true;
      code = code.replace('texel/_plugin/entry', opts.entry);
      code = `var preset = require(${JSON.stringify(presetFile)});\nvar result = ${code}\npreset(result);\n`;
      const header = '/** This module has been inserted by Texel to support optional sketches and presets. */';
      return `${header}\n${code}`.trim();
    } else {
      hasGeneratedEntry = false;
      return '';
    }
  }

  function patchFileWatch () {
    if (opts.mode !== 'development') return;
    process.nextTick(() => {
      // We need to monkey-patch this to force invalidate the generated
      // entry files. If at some point we replace watchify with Texel's
      // own file watching solution, we can get around this hack.
      const cache = bundler._options.cache;
      const watcher = bundler._watcher;
      const watchMap = {};
      if (typeof bundler._watcher !== 'function') {
        throw new Error(`It seems like Texel has somehow received an unexpected version of watchify in its node_modules; please file a new issue with Texel.`);
      }
      bundler._watcher = function (file, wOpts) {
        const result = watcher.call(bundler, file, wOpts);
        if (filesToWatch.includes(file)) {
          watchMap[file] = result;
        }
        if (file === opts.entry) {
          const onChange = () => {
            // When entry changes, delete other files associated with it
            // And also try to emit change events so watchify will un-load
            // their file watchers
            filesToWatch.forEach(other => {
              if (file === other) return;
              if (cache) delete cache[other];

              if (other in watchMap && watchMap[other]) {
                watchMap[other].emit('change');
              }
            });
          };
          result.on('change', onChange);
          const close = result.close;
          result.close = function () {
            result.removeListener('change', onChange);
            delete watchMap[file];
            return close.call(result);
          };
        }
        return result;
      };
    });
  }
};
