/** @jsx h */
require('./style.css');

const { h, render } = require('preact');
const exporter = require('texel/util/exporter');
const defined = require('defined');
const getContext = require('get-canvas-context');
const presetHeader = '[preset: "canvas"]\n';
const exampleSketch = `  export settings = { preset: 'print' };

  export default function () {
    return {
      draw: () => {}
    }
  }`;

import App from './ui/App';

module.exports = create;
function create (sketch) {
  if (!sketch) throw new Error(`It looks like you've exported a "settings" object but somehow your main sketch is not defined.`);
  if (!sketch.settings) throw new Error(`It looks like you've exported a "settings" object that is undefined or not truthy. Instead, try exporting an object with parameters.`);

  const settings = sketch.settings || {};

  const sketchFunc = sketch.default;
  if (typeof sketchFunc !== 'function') {
    console.warn(`${presetHeader}Your sketch file does not export a default function. It should look something like this:\n\n${exampleSketch}`);
  }

  const contextOpt = settings.context || '2d';
  let context;
  let isCanvas2D;
  if (typeof contextOpt === 'string') {
    context = getContext(contextOpt);
    if (!context) {
      throw new Error(`Error creating ${contextOpt} context`);
    }
    isCanvas2D = contextOpt === '2d';
  } else {
    context = contextOpt;
    isCanvas2D = typeof context.bindBuffer !== 'function' &&
        typeof context.scissor !== 'function';
  }

  const canvas = context.canvas;
  document.body.appendChild(canvas);

  const app = {
    canvas,
    context,
    width: 0,
    height: 0,
    pixelRatio: 1,
    resize,
    destroy,
    time: 0
  };

  resize();
  window.addEventListener('resize', resize);
  window.addEventListener('keydown', keydown);

  let sketchResult;
  if (typeof sketchFunc === 'function') {
    sketchResult = sketchFunc(app);
    if (!sketchResult) {
      console.warn(`${presetHeader}Your defualt function does not return an object. Try returning one, like this:\n\n${exampleSketch}`);
    }
  }
  window.requestAnimationFrame(animate);

  function animate (time) {
    app.time = time;
    window.requestAnimationFrame(animate);
    renderFrame();
  }

  function save () {
    setSize(true);
    renderFrame();
    const promise = exporter.saveCanvas(canvas);
    setSize(false);
    renderFrame();
    return promise;
  }

  function renderFrame () {
    if (isCanvas2D) {
      context.save();
      context.fillStyle = 'black';
      context.strokeStyle = 'black';
      context.lineWidth = 1;
      context.lineCap = 'butt';
      context.lineJoin = 'miter';
      context.miterLimit = 10;

      let scaleX = app.viewportPixelRatio * (app.viewportWidth / app.width);
      let scaleY = app.viewportPixelRatio * (app.viewportHeight / app.height);
      context.scale(scaleX, scaleY);
    }

    if (sketchResult && typeof sketchResult.draw === 'function') {
      sketchResult.draw(app);
    }

    if (isCanvas2D) {
      context.restore();
    }
  }

  function keydown (ev) {
    const cmdOrCtrl = ev.ctrlKey || ev.metaKey;
    if (ev.keyCode === 83 && cmdOrCtrl) {
      ev.preventDefault();
      save();
    }
  }

  function destroy () {
    if (canvas.parentNode) canvas.parentNode.removeChild(canvas);
    window.removeEventListener('resize', resize);
  }

  function resize () {
    setSize(false);
  }

  function setSize (matchSize = false) {
    const maxPixelRatio = defined(settings.maxPixelRatio, Infinity);
    app.width = defined(settings.width, window.innerWidth);
    app.height = defined(settings.height, window.innerHeight);
    app.pixelRatio = Math.min(maxPixelRatio, window.devicePixelRatio);

    const isLayout = (app.width !== window.innerWidth || app.height !== window.innerHeight);

    const aspect = app.width / app.height;
    const displayPadding = isLayout ? 40 : 0;
    const windowWidth = window.innerWidth;
    const windowHeight = window.innerHeight;
    const windowAspect = windowWidth / windowHeight;

    let width, height;
    if (windowAspect > aspect) {
      height = windowHeight - displayPadding * 2;
      width = height * aspect;
    } else {
      width = windowWidth - displayPadding * 2;
      height = width / aspect;
    }

    const matchPixelRatio = isLayout ? 1 : app.pixelRatio;
    let pixelRatio = matchSize ? matchPixelRatio : app.pixelRatio;
    let canvasWidth = matchSize ? app.width : width;
    let canvasHeight = matchSize ? app.height : height;
    const newWidth = Math.floor(canvasWidth * pixelRatio);
    const newHeight = Math.floor(canvasHeight * pixelRatio);
    if (canvas.width !== newWidth || canvas.height !== newHeight) {
      canvas.width = newWidth;
      canvas.height = newHeight;
    }
    canvas.style.width = `${width}px`;
    canvas.style.height = `${height}px`;
    app.viewportWidth = canvasWidth;
    app.viewportHeight = canvasHeight;
    app.viewportPixelRatio = pixelRatio;
  }

  function resizeSimple () {
    const maxPixelRatio = defined(settings.maxPixelRatio, Infinity);

    app.width = defined(settings.width, window.innerWidth);
    app.height = defined(settings.height, window.innerHeight);
    app.pixelRatio = Math.min(maxPixelRatio, window.devicePixelRatio);

    canvas.width = Math.floor(app.width * app.pixelRatio);
    canvas.height = Math.floor(app.height * app.pixelRatio);
    canvas.style.width = `${app.width}px`;
    canvas.style.height = `${app.height}px`;
    if (sketchResult && typeof sketchResult.onResize === 'function') {
      sketchResult.onResize();
    }
  }
}
