/** @jsx h */
import { h, Component } from 'preact';
import exporter from 'texel/util/exporter';
import defined from 'defined';

export default class Canvas extends Component {
  constructor (props) {
    super(props);
    this.state = {
      recording: false
    };

    this.highQuality = true;

    this.config = props;
    this.duration = defined(props.duration, 4);
    this.fps = defined(props.fps, 60);
    this.width = defined(props.width, window.innerWidth);
    this.height = defined(props.height, window.innerHeight);
    this.totalFrames = Math.ceil(this.fps * this.duration);
    this.frameInterval = 1000 / this.fps;

    this.viewportWidth = 0;
    this.viewportHeight = 0;
    this.time = 0;
    this.frame = 0;
    this.then = 0;
    this.raf = null;
  }

  componentDidMount () {
    this.canvas = this.base;
    this.context2d = this.canvas.getContext('2d');

    this.scene = null;

    this.resize();
    window.addEventListener('resize', this.resize);
    if (process.env.NODE_ENV === 'development') {
      window.addEventListener('keydown', this.keydown);
    }

    if (this.props.scene && typeof this.props.scene === 'function') {
      this.scene = this.props.scene({
        viewportWidth: this.viewportWidth,
        viewportHeight: this.viewportHeight,
        viewportPixelRatio: this.viewportPixelRatio,
        canvas: this.canvas,
        context: this.context2d,
        width: this.width,
        height: this.height
      });
    }

    // Start animation loop
    this.start();
  }

  componentWillUnmount () {
    this.stop();
    window.removeEventListener('resize', this.resize);
  }

  componentDidUpdate (oldProps) {
    if (this.props.recording !== oldProps.recording) {
      // Turn off animation loop
      this.stop();
      // Reset frame
      this.reset();

      if (this.props.recording && this.scene && typeof this.scene.onBeginRecord === 'function') {
        this.scene.onBeginRecord();
      }

      if (this.props.recording === 'gif') {
        // Start recording our GIF
        exporter.beginRecord({
          ...this.config,
          width: this.width,
          height: this.height
        }).then(() => this.record());
      } else if (this.props.recording) {
        this.record();
      } else {
        // Ensure we resize to window quality
        this.resize();
        // Turn on animation loop
        this.start();
      }
    }
  }

  stop () {
    if (this.raf === null) return;
    window.cancelAnimationFrame(this.raf);
    this.raf = null;
  }

  start () {
    this.raf = window.requestAnimationFrame(this.animate);
  }

  reset () {
    this.frame = 0;
    this.time = 0;
  }

  record () {
    const recording = this.props.recording;
    if (!recording || !this.base) return;
    // Set to output size
    this.resize();
    // Draw a single frame
    this.draw();
    if (this.scene && typeof this.scene.onRecordFrame === 'function') {
      this.scene.onRecordFrame();
    }
    // Save the canvas to a file
    const p = recording === 'gif'
      ? exporter.recordFrame(this.canvas)
      : exporter.saveCanvas(this.canvas, {
        ...this.config,
        frame: this.frame,
        totalFrames: Math.max(100, this.totalFrames)
      });

    // Indicate the current progress to UI
    this.props.onRecordProgress({ frame: this.frame, totalFrames: this.totalFrames });

    p.then(() => {
      // Step forward to next frame
      this.frame++;
      this.time += this.frameInterval / 1000;
      if (this.frame < this.totalFrames) {
        this.record();
      } else {
        if (this.scene && typeof this.scene.onEndRecord === 'function') {
          this.scene.onEndRecord();
        }
        if (recording === 'gif') exporter.endRecord();
        this.props.onRecordFinish();
      }
    });
  }

  keydown = (ev) => {
    const cmdOrCtrl = ev.ctrlKey || ev.metaKey;
    if (ev.keyCode === 83 && cmdOrCtrl) {
      ev.preventDefault();
      this.setSize(true);
      exporter.saveCanvas(this.canvas, {
        ...this.config
      }).then(result => {
        console.log('Image saved.');
      });
      this.setSize(false);
    }
  }

  resize = () => {
    this.setSize();
  }

  setSize (forceMatchSize = false) {
    const aspect = this.width / this.height;
    const displayPadding = 40;
    const windowWidth = window.innerWidth;
    const windowHeight = window.innerHeight;
    const windowAspect = windowWidth / windowHeight;

    let width, height;
    if (windowAspect > aspect) {
      height = windowHeight - displayPadding * 2;
      width = height * aspect;
    } else {
      width = windowWidth - displayPadding * 2;
      height = width / aspect;
    }

    let matchSize = this.highQuality === false || this.props.recording || forceMatchSize;
    let pixelRatio = matchSize ? 1 : window.devicePixelRatio;
    let canvasWidth = matchSize ? this.width : width;
    let canvasHeight = matchSize ? this.height : height;
    const newWidth = Math.floor(canvasWidth * pixelRatio);
    const newHeight = Math.floor(canvasHeight * pixelRatio);
    if (this.canvas.width !== newWidth || this.canvas.height !== newHeight) {
      this.canvas.width = newWidth;
      this.canvas.height = newHeight;
    }
    this.canvas.style.width = `${width}px`;
    this.canvas.style.height = `${height}px`;
    this.viewportWidth = canvasWidth;
    this.viewportHeight = canvasHeight;
    this.viewportPixelRatio = pixelRatio;
    this.pixelRatio = pixelRatio;

    this.draw();
  }

  animate = (now) => {
    if (this.props.recording) return; // Stop animation loop
    this.raf = window.requestAnimationFrame(this.animate);

    const dt = now - this.then;
    if (dt > this.frameInterval) {
      this.then = now - (dt % this.frameInterval);
      this.time += dt / 1000;
      this.draw();
      this.frame++;
      if (this.frame > this.totalFrames - 1) {
        this.reset();
        this.then = now;
      }
    }
  }

  draw = () => {
    const context = this.context2d;

    context.save();

    let scaleX = this.pixelRatio * (this.viewportWidth / this.width);
    let scaleY = this.pixelRatio * (this.viewportHeight / this.height);
    context.scale(scaleX, scaleY);

    context.lineJoin = 'round';
    context.lineCap = 'round';
    context.globalAlpha = 1;
    context.lineWidth = 1;
    context.fillStyle = 'white';
    context.fillRect(0, 0, this.width, this.height);
    context.fillStyle = 'black';
    context.strokeStyle = 'black';

    const progress = this.totalFrames === 0 ? 0 : this.frame / this.totalFrames;
    if (this.scene && typeof this.scene.draw === 'function') {
      this.scene.draw({
        viewportWidth: this.viewportWidth,
        viewportHeight: this.viewportHeight,
        viewportPixelRatio: this.viewportPixelRatio,
        canvas: this.canvas,
        context: this.context2d,
        width: this.width,
        height: this.height,
        frame: this.frame,
        time: this.time,
        progress
      });
    }

    context.restore();
  }

  render () {
    return <canvas />;
  }
}
