const getPackage = require('./get-package');
const { SimpleError } = require('./logger');
const chalk = require('chalk');
const path = require('path');
const createBabelify = require('../patched-modules/babelify');

// By default, JSX is tailored to React
const defaultPragma = 'React.createElement';

// Try to detect a few popular libraries out of the box
const altPragmaModules = [
  { name: 'preact', pragma: 'h' },
  { name: 'deku', pragma: 'h' },
  { name: 'hyperapp', pragma: 'h' },
  { name: 'vue', pragma: 'vue' },
  { name: 'react', pragma: defaultPragma }
];

module.exports = function (opts = {}) {
  const logger = opts.logger;
  return getPackage(opts.cwd).then(pkg => {
    let pragma = defaultPragma;
    if (typeof opts.jsx === 'string') {
      if (!opts.jsx) throw new SimpleError('No JSX pragma specified; should be a non-empty string, boolean, or null/undefined.');
      // User specifies a specific pragma
      pragma = opts.jsx;
    } else if (typeof opts.jsx === 'boolean') {
      // Use default or disable JSX entirely
      pragma = opts.jsx ? defaultPragma : false;
    } else if (pkg) {
      // Let texel infer pragma automatically
      const deps = [
        pkg.dependencies,
        pkg.devDependencies,
        pkg.optionalDependencies,
        pkg.peerDependencies
      ].filter(Boolean)
        .map(d => Object.keys(d))
        .reduce((a, b) => a.concat(b), []);

      const pragmaModules = altPragmaModules.filter(m => deps.includes(m.name));
      if (pragmaModules.length > 1) {
        logger.warn(`Your project depends on ${chalk.bold(pragmaModules[0].name)} and ${chalk.bold(pragmaModules[1].name)} which use different JSX pragmas. You may need to use the --jsx flag.\n\nThe default behaviour is to support React, like so:\n  texel start --jsx=React.createElement\n\nTo support Vue:\n  texel start --jsx=vue\n\nOr, to support preact, hyperapp, etc:\n  texel start --jsx=h`);
      } else if (pragmaModules.length === 1) {
        pragma = pragmaModules[0].pragma;
        // Maybe remove this log if it proves annoying or confusing for deku/hyperapp/preact crowd
        if (pragma !== defaultPragma) {
          logger.bullet(chalk.dim(`Inferring --jsx=${pragma} because of dependency on ${chalk.bold(pragmaModules[0].name)}`));
        }
      }
    }

    // TODO: It would be nice for users if there was a plugin that detected
    // when 'vue' and other modules are required() but --jsx is not set properly, i.e. in case
    // no package.json exists.
    let jsxPlugins = [];
    if (pragma === 'vue') {
      // Vue handles JSX differently than React

      logger.warn(`
Note that Vue's JSX transform is not yet supported.

Please see this issue for details and discussion:
  https://github.com/vuejs/babel-plugin-transform-vue-jsx/issues/127
      `.trim());

      // Use "h" pragma for now, since it sorta works with Vue
      pragma = 'h';

      // TODO: Support Vue, either by patching upstream module or forking it.
      // jsxPlugins = [ require.resolve('babel-plugin-transform-vue-jsx') ];
    }

    if (pragma) {
      jsxPlugins = [
        [ require.resolve('@babel/plugin-transform-react-jsx'), { pragma } ],
        require.resolve('@babel/plugin-syntax-jsx')
      ];
    }

    let useBuiltIns = false;
    if (opts.polyfill) {
      useBuiltIns = opts.optimize ? 'usage' : 'entry';
    }

    // Prepare the babel export extract plugin
    opts.extractExports = {
      filter: (file) => file === opts.entry,
      exports: {
        settings: {}
      }
    };

    const babelOpts = {
      cwd: opts.cwd,
      sourceMaps: opts.debug ? undefined : false
    };

    const shouldEnvify = !opts.node;
    const productionEnvify = shouldEnvify && opts.mode === 'production' ? [
      createBabelify.configure(Object.assign({}, babelOpts, {
        babelrc: false,
        // Optimize away certain files that don't need to be re-transformed
        passThroughFilter: (filename, tOpt) => {
          const dir = path.dirname(filename);
          const parts = path.relative(opts.cwd, dir).split(path.sep);
          if (parts[0] !== 'node_modules') {
            return false;
          }
          if (opts._noParseFiles.some(f => filename.endsWith(f))) {
            return false;
          }
          return true;
        },
        // If the file doesn't have "process.env", skip babelifying
        filter: (src, filename, opts) => {
          if (src.indexOf('process.env') === -1) return false;
          return true;
        },
        plugins: [
          [ require('../plugins/babel-process-env'), { env: opts.env } ]
        ]
      })),
      { global: true }
    ] : null;

    const babelify = createBabelify.configure(Object.assign({}, babelOpts, {
      presets: [
        [
          // TODO: "node" should be set to "current" if run with --node option?
          require.resolve('@babel/preset-env'), {
            useBuiltIns
          }
        ]
      ],
      passThroughFilter: (filename, tOpt) => {
        if (opts._noParseFiles.some(f => filename.endsWith(f))) {
          return false;
        }
        return true;
      },
      plugins: [
        // Remap certain import statements into require() statements to avoid any static
        // analysis issues when babel introduces inter-op for CommonJS exports. This is pretty ugly...!
        // Ideally there will be better static analysis tools in the future to handle interop weirdness.
        shouldEnvify ? [ require('../plugins/babel-process-env'), { env: opts.env } ] : null,
        [ require('babel-plugin-import-to-require'), {
          modules: [ 'fs', 'glslify', 'path', 'shader-reload' ]
        } ],
        // Include some of the most popular "experimental" features. This is supported
        // but subject to removal in a future version of texel.
        require('@babel/plugin-proposal-object-rest-spread'),
        require('@babel/plugin-proposal-async-generator-functions'),
        require('@babel/plugin-proposal-decorators'),
        // Loose is needed here until Stage 2 of decorators...
        [ require('@babel/plugin-proposal-class-properties'), { 'loose': true } ],
        require('../plugins/babel-require-resolve'),
        [ require('../plugins/babel-extract-exports'), opts.extractExports ]
      ].filter(Boolean).concat(jsxPlugins)
    }));
    // TODO: For glslify and brfs, it would be nice to remove unused 'path' and 'fs'
    // requires that might stick around after static analysis has completed. This can
    // reduce the bundle size somewhat significantly for small apps.
    return [
      babelify,
      require('glslify'),
      productionEnvify
    ].filter(Boolean);
  });
};
