const path = require('path');
const fs = require('fs');
const mkdirp = require('mkdirp');
const isBinaryFile = require('isbinaryfile');
const maxstache = require('maxstache');
const fsCopyFile = require('fs-copy-file');
const walker = require('./walker');
const DEFAULT_IGNORES = require('./ignore-files');

module.exports = copyFolderWithTemplate;
function copyFolderWithTemplate (input, output, opt = {}) {
  return mkdirAsync(output)
    .then(config => templateCopy(input, output, opt));
}

function mkdirAsync (file) {
  return new Promise((resolve, reject) => {
    mkdirp(file, err => {
      if (err) reject(err);
      else resolve();
    });
  });
}

function swapDir (inFile, input, output) {
  return path.resolve(output, path.relative(input, inFile));
}

function isFilterMatch (dir, filters = []) {
  const name = path.basename(dir);
  return filters.some(filter => typeof filter === 'string' ? name === filter : filter.test(name));
}

// Check if the folder is a top-level node_modules folder (i.e. ignore it)
function isNodeModules (dir, inputFolder) {
  const parts = path.relative(inputFolder, dir).split(path.sep);
  const mod = parts[0] === 'node_modules';
  return mod;
}

function copyFile (inFile, input, output, templateData, templateIgnores) {
  return new Promise((resolve, reject) => {
    // see if we should template or not
    const outFile = swapDir(inFile, input, output);
    if (isFilterMatch(inFile, templateIgnores) || isBinaryFile.sync(inFile)) {
      // binary or template-ignored file, just copy it
      console.log(path.basename(inFile), path.basename(outFile));
      fsCopyFile(inFile, outFile, err => {
        if (err) reject(err);
        else resolve();
      });
    } else {
      // textual file, template it
      fs.readFile(inFile, 'utf8', (err, data) => {
        if (err) reject(err);
        data = maxstache(data, templateData);
        fs.writeFile(outFile, data, err => {
          if (err) reject(err);
          else resolve();
        });
      });
    }
  });
}

function templateCopy (input, output, opt = {}) {
  const ignores = DEFAULT_IGNORES.concat(opt.ignore || []).filter(Boolean);
  const templateIgnores = [].concat(opt.templateIgnore || []).filter(Boolean);
  const data = opt.data || {};
  return walker({
    filterFolders: params => {
      return !isNodeModules(params.fullPath, params.root) && !isFilterMatch(params.fullPath, ignores);
    },
    filterFiles: params => {
      return !isFilterMatch(params.fullPath, ignores);
    },
    path: input
  }, (params) => {
    if (params.directory) {
      const dir = swapDir(params.fullPath, input, output);
      return mkdirAsync(dir);
    } else {
      return copyFile(params.fullPath, input, output, data, templateIgnores);
    }
  });
}
