// During development, this module can be accessed like so:
//   import { saveCanvas } from 'texel/util/exporter'

// It will get optimized into no-op functions on bundle.

function post (url, opts = {}) {
  if (!('fetch' in window)) {
    return Promise.reject(new Error('It looks like you are on an older browser without the fetch() API. The texel/export features rely on newer browsers. Try using Chrome or FireFox.'));
  }
  return window.fetch(url, {
    method: 'POST',
    headers: {
      'Accept': 'application/json',
      'Content-Type': 'application/json'
    },
    body: JSON.stringify(opts)
  }).then(res => {
    if (res.status >= 200 && res.status < 300) {
      return res;
    } else if (res.status === 404) {
      console.error('Somehow texel/export received a 404 status when trying to save the file. Make sure you are running this code through texel command-line or application, and in development mode.');
    } else {
      throw new Error(res.statusText);
    }
  });
}

module.exports.saveFile = function (data, opts = {}) {
  return post('/texel/export/saveFile', {
    output: opts.output,
    extension: opts.extension,
    file: opts.file,
    frame: opts.frame,
    totalFrames: opts.totalFrames,
    data: data
  });
};

module.exports.saveCanvas = function (canvas, opts = {}) {
  const { type, extension } = getImageType(opts.format);
  const base64 = getBase64(canvas, type, opts.quality);
  return module.exports.saveFile(base64, Object.assign({ extension }, opts));
};

module.exports.beginRecord = function (opts = {}) {
  if (typeof opts.width !== 'number' || typeof opts.height !== 'number') throw new TypeError('Must specify a width and height for the recording');
  return post('/texel/export/beginRecord', Object.assign({}, opts, { extension: '.gif' }));
};

module.exports.recordFrame = function (canvas) {
  // Send PNG base64 data, allow server to decode it and add RGBA to GIF encoder
  const { type, extension } = getImageType('png');
  const base64 = getBase64(canvas, type);
  return post('/texel/export/recordFrame', {
    extension,
    data: base64
  });
};

module.exports.endRecord = function (opts = {}) {
  return post('/texel/export/endRecord', opts);
};

function getBase64 (canvas, type, quality) {
  return canvas.toDataURL(type, quality).slice(`data:${type};base64,`.length);
}

function getImageType (format = 'png') {
  switch (format) {
    case 'jpeg':
    case 'jpg':
      return { extension: '.jpg', type: 'image/jpeg' };
    case 'webp':
      return { extension: '.webp', type: 'image/webp' };
    case 'png':
    default:
      return { extension: '.png', type: 'image/png' };
  }
}
